/*
	File: PlaneWM.C
	Component of the Stochastic Discrimination Utilities
	Copyright (C) 2002, David C. Lambert <dcl@panix.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation, Inc.,
	59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include <math.h>
#include <string.h>
#include "PlaneWM.H"

// Uncomment the '#define' below if you want the hyperplanes' normal vectors
// to be random, rather than parallel to the vector connecting the
// randomly chosen data points (the difference vector).  I put this in
// because it's cool, but I have no idea how it affects performance.

//#define RANDOM_VECT

WeakModel *
PlaneWMFactory(WMSet &set, int nDims)
{
	return (new PlaneWM(nDims));
}

WeakModel *
PlaneWMCopier(const WeakModel &rhs)
{
	return (new PlaneWM((PlaneWM &)rhs));
}

WeakModel *
PlaneWMLoader(int nDims, FILE *dumpFile)
{
	return (new PlaneWM(nDims, dumpFile));
}

void
PlaneWMParamUpdate()
{
}

#ifdef RANDOM_VECT

// use random vector for projections, rather than difference
// vector

PlaneWM::PlaneWM(int nDims)
		: WeakModel(nDims), _k(0.0), _v(new double[_setDims])
{
	double		kIn = 0.0;
	double		kOut = 0.0;
	double		scale = 0.0;
	double		offset = 0.0;
	double		*inClassPoint = WMSet::RandomPositiveDataPoint();
	double		*outClassPoint = WMSet::RandomNegativeDataPoint();

	if (_setDims < _nDims) {
		for(int i=0; i < _setDims; i++) {
			int		j = _inSet[i];
			_v[i] = 2.0 * drand48() - 1.0;
		}
	}
	else {
		for(int i=0; i < _nDims; i++)
			_v[i] = 2.0 * drand48() - 1.0;
	}
	kIn = _dotProductWithVect(inClassPoint);
	kOut = _dotProductWithVect(outClassPoint);

	scale = (kIn - kOut)/2.0;
	offset = kOut;

	double	val = drand48() * (WMSet::RhoMax - WMSet::RhoMin) + WMSet::RhoMin; 
	_k = scale * (val + 1.0) + offset;
}

#else

PlaneWM::PlaneWM(int nDims)
		: WeakModel(nDims), _k(0.0), _v(new double[_setDims])
{
	double		kIn = 0.0;
	double		kOut = 0.0;
	double		scale = 0.0;
	double		offset = 0.0;
	double		*inClassPoint = WMSet::RandomPositiveDataPoint();
	double		*outClassPoint = WMSet::RandomNegativeDataPoint();

	if (_setDims < _nDims) {
		for(int i=0; i < _setDims; i++) {
			int		j = _inSet[i];
			double	diff = inClassPoint[j] - outClassPoint[j];
			_v[i] = diff;
		}
	}
	else {
		for(int i=0; i < _nDims; i++) {
			double	diff = inClassPoint[i] - outClassPoint[i];
			_v[i] = diff;
		}
	}
	kIn = _dotProductWithVect(inClassPoint);
	kOut = _dotProductWithVect(outClassPoint);

	scale = (kIn - kOut)/2.0;
	offset = kOut;

	double	val = drand48() * (WMSet::RhoMax - WMSet::RhoMin) + WMSet::RhoMin; 
	_k = scale * (val + 1.0) + offset;
}
#endif


PlaneWM::PlaneWM(const PlaneWM &rhs)
		: WeakModel(rhs), _k(rhs._k), _v(new double[rhs._setDims])
{
	memcpy(_v, rhs._v, sizeof(double)*_setDims);
}

PlaneWM::PlaneWM(int nDims, FILE *dumpFile)
		: WeakModel(nDims, dumpFile), _k(0.0), _v(new double [_setDims])
{
	fscanf(dumpFile, " { %lf ", &_k);
	for(int i = 0; i < _setDims; i++)
		fscanf(dumpFile, "%lf ", &_v[i]);
	fscanf(dumpFile, "} ");
}

double
PlaneWM::_dotProductWithVect(double *p)
{
	double		dotProduct = 0.0;

	if (_setDims < _nDims) {
		for(int i = 0; i < _setDims; i++) {
			int		j = _inSet[i];
			dotProduct += p[j] * _v[i];
		}
	}
	else {
		for(int i = 0; i < _nDims; i++)
			dotProduct += p[i] * _v[i];
	}
	return dotProduct;
}

int
PlaneWM::containsPoint(double *p)
{
	return (_dotProductWithVect(p) > _k);
}


void
PlaneWM::writeToFile(FILE *dumpFile)
{
	WeakModel::writeToFile(dumpFile);
	fprintf(dumpFile, " { %.5e ", _k);
	for(int i = 0; i < _setDims; i++)
		fprintf(dumpFile, "%.5e ", _v[i]);
	fprintf(dumpFile, "} ");
}

