/*
	File: RectWM.C
	Component of the Stochastic Discrimination Utilities
	Copyright (C) 2002, David C. Lambert <dcl@panix.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation, Inc.,
	59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include "RectWM.H"

WeakModel *
RectWMFactory(WMSet &set, int nDims)
{
	return (new RectWM(nDims));
}

WeakModel *
RectWMCopier(const WeakModel &rhs)
{
	return (new RectWM((RectWM &)rhs));
}

WeakModel *
RectWMLoader(int nDims, FILE *dumpFile)
{
	return (new RectWM(nDims, dumpFile));
}

void
RectWMParamUpdate()
{
	if (WMSet::RhoMax > WMSet::RhoMin) WMSet::RhoMax *= WMSet::RhoDecay;
	//fprintf(stderr, "RhoMax = %.3f\n", RectWM::RhoMax);
}


void
RectWM::_setBoundsUsing(double spanMultiplier, DPair &rectBounds, DPair &dataBounds, double rmid)
{
	double	minB = dataBounds.v0;
	double	maxB = dataBounds.v1;
	double	span = maxB-minB;
	//double	spanT = span * WMSet::Tau;
	//double	hiLim = minB + spanT;
	//double	loLim = maxB - spanT;
	double	spanT = span * (WMSet::Tau - 1.0);
	double	hiLim = maxB + spanT;
	double	loLim = minB - spanT;
	double	width = (spanMultiplier * span) / 2.0;
	double	upper = rmid + width;
	double	lower = rmid - width;

	upper = (upper < hiLim) ? upper : hiLim;
	lower = (lower > loLim) ? lower : loLim;

	rectBounds.v0 = lower;
	rectBounds.v1 = upper;
}


RectWM::RectWM(int nDims)
		: WeakModel(nDims), _bounds(new DPair[_setDims])
{
	double		rhoDiff = WMSet::RhoMax - WMSet::RhoMin;
	double		spanMultiplier = drand48() * (WMSet::RhoMax - WMSet::RhoMin);
	double		*point = WMSet::RandomDataPoint();

	if (_setDims < _nDims) {
		for(int i = 0; i < _setDims; i++) {
			int		j = _inSet[i];
			_setBoundsUsing(spanMultiplier, _bounds[i], WMSet::DataBounds(j), point[j]);
		}
	}
	else {
		for(int i = 0; i < _nDims; i++)
			_setBoundsUsing(spanMultiplier, _bounds[i], WMSet::DataBounds(i), point[i]);
	}
}

RectWM::RectWM(const RectWM &rhs)
		: WeakModel(rhs), _bounds(new DPair[rhs._setDims])
{
	for(int i=0;i<_setDims;i++)
		_bounds[i] = rhs._bounds[i];
}

RectWM::RectWM(int nDims, FILE *dumpFile)
		: WeakModel(nDims, dumpFile), _bounds(0)
{
	fscanf(dumpFile, "{ ");
	_bounds = new DPair[_setDims];
	for(int i = 0; i < _setDims; i++) {
		DPair	&dp = _bounds[i];
		fscanf(dumpFile, " %lf/%lf ", &dp.v0, &dp.v1);
	}
	fscanf(dumpFile, "} ");
}

int
RectWM::containsPoint(double *p)
{
	if (_setDims < _nDims) {
		for(int i = 0; i < _setDims; i++) {
			int		j = _inSet[i];
			DPair	&dp = _bounds[i];
			double	pj = p[j];
			if (pj < dp.v0 || pj > dp.v1) return 0;
		}
	}
	else {
		for(int i = 0; i < _nDims; i++) {
			DPair	&dp = _bounds[i];
			double	pi = p[i];
			if (pi < dp.v0 || pi > dp.v1) return 0;
		}
	}
}


void
RectWM::writeToFile(FILE *dumpFile)
{
	WeakModel::writeToFile(dumpFile);
	fprintf(dumpFile, " { ");
	for(int i = 0; i < _setDims; i++) {
		DPair	&dp = _bounds[i];
		fprintf(dumpFile, "%.5e/%.5e ", dp.v0, dp.v1);
	}
	fprintf(dumpFile, "} ");
}

