/*
	File: SD.C
	Component of the Stochastic Discrimination Utilities
	Copyright (C) 2002, David C. Lambert <dcl@panix.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation, Inc.,
	59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include "SD.H"

// uncomment the '#define' below if you want beta to be maximum ratio
// of negative to positive coverage frequencies that constitutes an
// enriched set, rather than the minimum difference in their raw scores.
//
// [more intuitive for me, doesn't seem to have much impact on performance]
//
// NOTE: if you do this, you will reverse the relationship of beta value
//       to the 'difficulty' of the enrichment test.  In the default configuration,
//       increasing beta makes the enrichment test *harder* for a candidate set
//       to pass; if DCL_BETA is defined, a larger beta makes the enrichment
//       test *easier* for a candidate set to pass.

//#define DCL_BETA

SD::SD(int nData, int nDims, int maxSets, double beta, double kappa, double tau,
		 double rhoMin, double rhoMax, double rhoDecay, double **data, int *classes, DPair *bounds)
		:_nDims(nDims), _nData(nData), _maxSets(maxSets), _nSets(0), _nPositive(0), _nNegative(0),
		 _beta(beta), _kappa(kappa), _tau(tau), _rhoMin(rhoMin), _rhoMax(rhoMax), _rhoDecay(rhoDecay),
		 _covered(new int [nData]), _counters(new double [nData]), _frequencies(new double [nData]),
		 _fracPos((double *)0), _fracNeg((double *)0), _classes(classes), _data(data), _bounds(bounds),
		 _sets((WMSet **)0), _avgPosNumber(0.0), _avgNegNumber(0.0), _posIndices(0), _negIndices(0)
{
	int		i = 0;
	int		np = 0;
	int		nn = 0;

	_sets = new WMSet * [_maxSets];
	for(i = 0; i < _maxSets; i++)
		_sets[i] = (WMSet *)0;

	for(i = 0; i < _nData; i++) {
		int		c = _classes[i];
		if (c == 1) _nPositive++;
		else _nNegative++;
		_covered[i] = 0;
		_counters[i] = _frequencies[i] = 0.0;
	}

	_posIndices = new int [(int)_nPositive];
	_negIndices = new int [(int)_nNegative];

	for(i = 0; i < _nData; i++) {
		int		c = _classes[i];
		if (c == 1)
			_posIndices[np++] = i;
		else
			_negIndices[nn++] = i;
	}
}

SD::SD(FILE *dumpFile)
		: _nDims(0), _nData(0), _maxSets(0), _nSets(0), _nPositive(0), _nNegative(0),  _beta(0), _kappa(0), _tau(0),
		  _rhoMin(0), _rhoMax(0), _rhoDecay(0), _covered((int *)0), _counters((double *)0),
		  _frequencies((double *)0), _fracPos((double *)0), _fracNeg((double *)0), _classes((int *)0),
		  _data((double **)0), _bounds((DPair *)0), _sets((WMSet **)0), _avgPosNumber(0.0), _avgNegNumber(0.0)
{
	fscanf(dumpFile, "%d %d", &_nSets, &_nDims);

	_sets = new WMSet * [_nSets];
	for(int i = 0; i < _nSets; i++) {
		int		j;
		fscanf(dumpFile, "%d", &j);
		_sets[i] = new WMSet(_nDims, dumpFile);
	}
}

void
SD::writeToFile(FILE *dumpFile)
{
	fprintf(dumpFile, "%d %d\n", _nSets, _nDims);
	for(int i = 0; i < _nSets; i++) {
		WMSet		*set = _sets[i];
		fprintf(dumpFile, "%d ", i);
		set->writeToFile(dumpFile);
	}
}

int
SD::classifyPoint(double *p, double &avg, double thresh)
{
	double		v = 0;

	for(int i = 0; i < _nSets; i++) {
		WMSet		*set = _sets[i];
		double	fracPos = set->fracPos();
		double	fracNeg = set->fracNeg();
		v += (set->containsPoint(p) - fracNeg)/(fracPos - fracNeg);
	}
	avg = v/_nSets;
	return ((avg >= thresh) ? 1 : 0);
}

double
SD::getAccuracy(int nData, double **data, int *classes, int doPrint, double thresh)
{
	int		i;
	int		j;
	double	avg;
	double	nRight = 0.0;
	int		rights[2][2];

	if (doPrint) {
		for(i=0; i < 2; i++) {
			for(j=0; j < 2; j++)
				rights[i][j] = 0;
		}
		fprintf(stderr, "\nindex       score  target output\n");
	}

	for(i = 0; i < nData; i++) {
		int		classID = classifyPoint(data[i], avg, thresh);
		if (classes[i] == classID) nRight++;
		rights[classes[i]][classID]++;
		if (doPrint) fprintf(stderr, "%010d %6.3f %6d %6d\n", i, avg, classes[i], classID);
	}

	if (doPrint) {
		fprintf(stderr, "\n               0     1  <- classified as\n");
		fprintf(stderr, "targets:");
		for(i = 0; i < 2; i++) {
			if (i > 0)
				fprintf(stderr, "        ");
			fprintf(stderr, "%2d ", i);
			for(j = 0; j < 2; j++)
				fprintf(stderr, "%5d ", rights[i][j]);
			fprintf(stderr, "\n");
		}
	}
	return nRight/nData;
}

double
SD::getAccuracy(int dump)
{
	return getAccuracy(_nData, _data, _classes, dump);
}

void
SD::getDistValues(double &minPos, double &maxPos, double &posAvg, double &posSD,
				   double &minNeg, double &maxNeg, double &negAvg, double &negSD)
{
	int			i;
	double		ssPosAvg = 0.0;
	double		ssNegAvg = 0.0;

	minPos = minNeg = 1.0e99;
	maxPos = maxNeg = -1.0e99;

	for(i = 0; i < _nData; i++) {
		int			classID = _classes[i];
		double		freq = _frequencies[i];

		if (classID == 1) {
			minPos = (minPos < freq) ? minPos : freq;
			maxPos = (maxPos > freq) ? maxPos : freq;
			posAvg += freq;
		}
		else {
			minNeg = (minNeg < freq) ? minNeg : freq;
			maxNeg = (maxNeg > freq) ? maxNeg : freq;
			negAvg += freq;
		}
	}
	posAvg /= _nPositive;
	negAvg /= _nNegative;

	for(i = 0; i < _nData; i++) {
		int			classID = _classes[i];
		double		freq = _frequencies[i];

		if (classID == 1)
			ssPosAvg += pow((posAvg-freq), 2.0);
		else
			ssNegAvg += pow((negAvg-freq), 2.0);
	}

	posSD = sqrt(ssPosAvg/_nPositive);
	negSD = sqrt(ssNegAvg/_nNegative);
}

void
SD::calculateCoverages(WMSet &set)
{
	double	nPos = 0.0;
	double	nNeg = 0.0;

	for(int i = 0; i < _nData; i++) {
		int			classID = _classes[i];
		double		*point = _data[i];
		int			covered = set.containsPoint(point);

		_covered[i] = 0;
		if (covered == 1)	{
			_covered[i] = 1;
			(classID == 1) ? nPos++ : nNeg++;
		}
	}
	set.fracPos(nPos/_nPositive);
	set.fracNeg(nNeg/_nNegative);
}


void
SD::testUniformity(WMSet &set, double &avgPosNumber, double &avgNegNumber)
{
	avgPosNumber = 0.0;
	avgNegNumber = 0.0;

	if (_nSets == 0) return;

	double		nPosPoints = 0.0;
	double		nNegPoints = 0.0;

	for(int i = 0; i < _nData; i++) {
		double		*point = _data[i];
		int			classID = _classes[i];
		int			covered = _covered[i];
		double		counter = _counters[i];

		if (covered == 1) {
			if (classID == 1) {
				nPosPoints++;
				avgPosNumber += counter;
			}
			else {
				nNegPoints++;
				avgNegNumber += counter;
			}
		}
	}

	if (nPosPoints > 0.0) avgPosNumber /= nPosPoints;
	if (nNegPoints > 0.0) avgNegNumber /= nNegPoints;
}


void
SD::updateVariables(WMSet &set)
{
	double	fracNeg = set.fracNeg();
	double	fracPos = set.fracPos();
	double	diff = fracPos - fracNeg;

	_avgPosNumber = _avgNegNumber = 0.0;

	for(int i = 0; i < _nData; i++) {
		int			classID = _classes[i];
		double		freq = 0.0;
		double		covered = (double)_covered[i];
		double		counter = _counters[i];
		double		*point = _data[i];

		counter += (covered - fracNeg)/diff;
		freq = counter/_nSets;

		_counters[i] = counter;
		_frequencies[i] = freq;

		if (classID == 1)
			_avgPosNumber += counter;
		else
			_avgNegNumber += counter;
	}
	_avgPosNumber /= _nPositive;
	_avgNegNumber /= _nNegative;
}


int
SD::addSubset(WMSet &newSet, int &noEnrichment, int &noUniformity, int skipUniformity)
{
	int		doAdd = 0;
	double	diff = 0.0;
	double	fracPos = 0.0;
	double	fracNeg = 0.0;
	double	avgPosNumber = 0.0;
	double	avgNegNumber = 0.0;

	noUniformity = 0;
	noEnrichment = 1;

	calculateCoverages(newSet);
	fracPos = newSet.fracPos(); fracNeg = newSet.fracNeg();

#ifdef DCL_BETA
	// make beta be the maximum ratio of fracNeg to fracPos
	// that constitutes an enriched set
	if ((fracPos > 0.0) && (fracNeg/fracPos < _beta))	{
#else
	if ((fracPos-fracNeg) > _beta) {
#endif
		noEnrichment = 0;
		noUniformity = 1;
		if (_nSets == 0 || skipUniformity) {
			doAdd = 1;
			noUniformity = 0;
		}
		else {
			testUniformity(newSet, avgPosNumber, avgNegNumber);
			if (avgPosNumber < _avgPosNumber) {
				doAdd = 1;
				noUniformity = 0;
			}
		}
	}
	return doAdd;
}

void
SD::addSubset(int tries)
{
	int		noEnrichment;
	int		noUniformity;
	int		nonUniform = 0;
	int		nonEnriched = 0;
	int		originalTries = tries;
	double	fracPos = 0.0;
	double	fracNeg = 0.0;
	WMSet	*newSet = (WMSet *)0;

	while(tries > 0) {
		int		nModels = (_kappa <= 1.0) ? 1 : (int)(1.0 + _kappa * drand48());
		newSet = new WMSet(_nDims, nModels);

		if (addSubset(*newSet, noEnrichment, noUniformity, 0))
			break;

		nonUniform += noUniformity;
		nonEnriched += noEnrichment;

		tries--;

		delete newSet;
	}

	if (tries <= 0) {
		tries = originalTries;
		while(tries > 0) {
			int		nModels = (_kappa <= 1.0) ? 1 : (int)(1.0 + _kappa * drand48());
			newSet = new WMSet(_nDims, nModels);

			if (addSubset(*newSet, noEnrichment, noUniformity, 1))
				break;

			nonUniform++;
			nonEnriched += noEnrichment;

			tries--;

			delete newSet;
		}
	}

	takeNewSet(newSet);

#if 0
	while((tries == -1) || tries > 0) {
		int		nModels = (_kappa <= 1.0) ? 1 : (int)(1.0 + _kappa * drand48());
		newSet = new WMSet(_nDims, nModels);

		if (addSubset(*newSet, noEnrichment, noUniformity, 0))
			break;

		nonUniform += noUniformity;
		nonEnriched += noEnrichment;

		if (tries != -1) tries--;

		delete newSet;
	}
#endif
}


void
SD::takeNewSet(WMSet *newSet)
{
	_sets[_nSets] = newSet;
	_nSets++;

	updateVariables(*newSet);

	WMSet::WMParamUpdate();
}
