/*
    File: SDTrain.C
    Component of the Stochastic Discrimination Utilities
    Copyright (C) 2002, David C. Lambert <dcl@panix.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include <string.h>
#include "SD.H"
#include "SDUtil.H"
#include "RectWM.H"
#include "HammWM.H"

typedef struct {
	// model index 0=Rect, 1=Hamm, 2=Plane
	int		modelType;

	// generate accuracy report when done
	int		dumpAccuracy;

	// fraction of input features to use when generating models
	double	subsetThresh;

	// test set rows
	int		nTest;

	// training set rows
	int		nTrain;

	// dimensionality of input vector
	int		nInputs;

	// maximum number of weak classifiers
	int		maxSets;

	// number of target classes
	int		nClasses;

	// display after 'dispInterval' classifier allocations
	int		dispInterval;

	// random number seed
	long	seed;

	// bounds multiplier for classifier rects
	double	tau;

	// enrichment test: (fracPos > 0 && fracNeg/fracPos < beta)
	double	beta;

	// make no more than kappa rectangles/classifier
	double	kappa;

	// minimum width as a function of span in single dimension
	double	rhoMin;

	// maximum width as a function of span in single dimension
	double	rhoMax;

	// rhoMax *= rhoDecay after every set added
	double	rhoDecay;
	
	// data file names
	char	dumpFileName[256];
	char	testFileName[256];
	char	trainFileName[256];
} SDArgs;

void
usage()
{
	fprintf(stderr, "\n");
	fprintf(stderr, "Usage: SDTrain -F trainFile -N nTrain -f testFile -n nTest -i nInputs [ -C nClasses][-M maxSets]\n");
	fprintf(stderr, "                [-I dispInterval][-b beta][-k kappa][-R rhoMax][-r rhoMin][-D rhoDecay][-s seed]\n");
	fprintf(stderr, "                [-t tau][-o dumpFileName][-m modelType][-d]\n");
	fprintf(stderr, "Defaults:\n");
	fprintf(stderr, "  nClasses = 2\n");
	fprintf(stderr, "  maxSets = 100\n");
	fprintf(stderr, "  dispInterval = 10\n");
	fprintf(stderr, "  beta = 0.1\n");
	fprintf(stderr, "  kappa = 1\n");
	fprintf(stderr, "  rhoMax = 0.125\n");
	fprintf(stderr, "  rhoMin = 0.005\n");
	fprintf(stderr, "  rhoDecay = 0.99990\n");
	fprintf(stderr, "  tau = 1.1\n");
	fprintf(stderr, "  seed = time(0)\n");
	fprintf(stderr, "  modelType = 0 (RectWM)\n");
	fprintf(stderr, "  subsetThresh = 1.0\n");
	fprintf(stderr, "  dumpAccuracy = 0\n");
	fprintf(stderr, "\n");
}

void
parseCmdLine(int argc, char **argv, SDArgs &args)
{
	int			c;
	extern char	*optarg;
	extern int	optind, opterr, optopt;

	args.nTest = 0;
	args.nTrain = 0;
	args.nInputs = 0;
	args.nClasses = 0;
	args.maxSets = 100;
	args.dispInterval = 10;
	args.seed = time(0);
	args.beta = 0.1;
	args.kappa = 1.0;
	args.tau = 1.1;
	args.rhoMin = 0.005;
	args.rhoMax = .125;
	args.rhoDecay = 0.99990;
	args.modelType = WMSet::HRECT;
	args.dumpAccuracy = 0;
	args.subsetThresh = 1.0;

	strcpy(args.testFileName, "/dev/null");
	strcpy(args.dumpFileName, "/dev/null");
	strcpy(args.trainFileName, "/dev/null");

	while((c = getopt(argc, argv, "o:n:N:i:f:F:b:r:R:dD:k:t:M:I:s:C:o:m:T:")) != -1) {
		switch(c) {
		  case 'T':
			  args.subsetThresh = atof(optarg);
			  break;
		  case 'm':
			  args.modelType = atoi(optarg);
			  break;
		  case 'M':
			  args.maxSets = atoi(optarg);
			  break;
		  case 'I':
			  args.dispInterval = atoi(optarg);
			  break;
		  case 'N':
			  args.nTrain = atoi(optarg);
			  break;
		  case 'n':
			  args.nTest = atoi(optarg);
			  break;
		  case 'F':
			  sprintf(args.trainFileName, "%.255s", optarg);
			  break;
		  case 'f':
			  sprintf(args.testFileName, "%.255s", optarg);
			  break;
		  case 'i':
			  args.nInputs = atoi(optarg);
			  break;
		  case 'b':
			  args.beta = atof(optarg);
			  break;
		  case 'k':
			  args.kappa = atof(optarg);
			  break;
		  case 'r':
			  args.rhoMin = atof(optarg);
			  break;
		  case 'R':
			  args.rhoMax = atof(optarg);
			  break;
		  case 'D':
			  args.rhoDecay = atof(optarg);
			  break;
		  case 't':
			  args.tau = atof(optarg);
			  break;
		  case 's':
			  args.seed = atol(optarg);
			  break;
		  case 'C':
			  args.nClasses = atoi(optarg);
			  break;
		  case 'o':
			  sprintf(args.dumpFileName, "%.255s", optarg);
			  break;
		  case 'd':
			  args.dumpAccuracy = 1;
			  break;
		  case '?':
			  usage();
			  exit(1);
		}
	}
	if (args.nTrain == 0 || args.nInputs == 0 || !strcmp(args.trainFileName, "/dev/null")) {
		usage();
		exit(1);
	}

	srand48(args.seed);
	fprintf(stderr, "seed = %d\n", args.seed);
}

int
main(int argc, char **argv)
{
	int		i;
	int		j;
	int		k;
	SDArgs	args;

	parseCmdLine(argc, argv, args);

	int			*trainClasses = new int [args.nTrain];
	double		**trainData = new double * [args.nTrain];
	for(i = 0; i < args.nTrain; i++) trainData[i] = new double [args.nInputs];

	int			*testClasses = new int [args.nTest];
	double		**testData = new double * [args.nTest];
	for(i = 0; i < args.nTest; i++) testData[i] = new double [args.nInputs];

	DPair		*trainBounds = new DPair [args.nInputs];
	for(i = 0; i < args.nInputs; i++) {
		trainBounds[i].v0 = 1.0e99;
		trainBounds[i].v1 = -1.0e99;
	}
		
	loadData(args.trainFileName, args.nTrain, args.nInputs, trainData, trainClasses, trainBounds);
	loadData(args.testFileName, args.nTest, args.nInputs, testData, testClasses, NULL);

	fprintf(stderr, "data load complete\n");

	if (!WMSet::Initialize(args.nTrain, trainData, trainClasses, trainBounds, (WMSet::TWMType)args.modelType))	{
		fprintf(stderr, "invalid model type %d selected\n", args.modelType);
		exit(1);
	}

	WMSet::Tau = args.tau;
	WMSet::RhoMin = args.rhoMin;
	WMSet::RhoMax = args.rhoMax;
	WMSet::RhoDecay = args.rhoDecay;
	WMSet::SubsetThresh = 1.0-args.subsetThresh;

	if (args.nClasses == 2) {
		SD		sd(args.nTrain, args.nInputs, args.maxSets, args.beta, args.kappa, args.tau,
				   args.rhoMin, args.rhoMax, args.rhoDecay, trainData, trainClasses, trainBounds);

		learnBinaryClassifier(sd, testData, testClasses, args.nTest, args.dispInterval, args.dumpAccuracy);

		if (strcmp(args.dumpFileName, "/dev/null"))
			dumpBinaryClassifier(sd, args.dumpFileName);
	}
	else {
		int				**testClassArray = new int * [args.nClasses];
		int				**trainClassArray = new int * [args.nClasses];
		SD				**sdArray = new SD * [args.nClasses];

		for(i = 0; i < args.nClasses; i++) {
			int		*testC = new int [args.nTest];
			int		*trainC = new int [args.nTrain];

			testClassArray[i] = testC;
			for(j = 0; j < args.nTest; j++) testC[j] = (testClasses[j] == i) ? 1 : 0;

			trainClassArray[i] = trainC;
			for(j = 0; j < args.nTrain; j++) trainC[j] = (trainClasses[j] == i) ? 1 : 0;

			sdArray[i] = new SD(args.nTrain, args.nInputs, args.maxSets, args.beta, args.kappa, args.tau,
								args.rhoMin, args.rhoMax, args.rhoDecay, trainData, trainC, trainBounds);
		}
		learnMultipleClassifier(sdArray, args.nClasses, args.dispInterval,
								args.nTrain, trainData, trainClasses, trainClassArray,
								args.nTest, testData, testClasses, testClassArray, args.dumpAccuracy);

		if (strcmp(args.dumpFileName, "/dev/null"))
			dumpMultipleClassifiers(sdArray, args.nClasses, args.dumpFileName);
	}
	exit(0);
}
