/*
    File: SDUtil.C
    Component of the Stochastic Discrimination Utilities
    Copyright (C) 2002, David C. Lambert <dcl@panix.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include "SDUtil.H"

SD *
loadBinaryClassifier(const char *dumpFileName)
{
	FILE	*dumpFile = fopen(dumpFileName, "r");

	SD		*loadedSD = new SD(dumpFile);

	fclose(dumpFile);

	return loadedSD;
}

void
learnBinaryClassifier(SD &sd, double **testData, int *testClasses, int nTest, int dispInterval, int dumpAccuracy)
{
	int		maxSets = sd.maxSets();
	double	posSD = 0.0;
	double	negSD = 0.0;
	double	minPos = 0.0;
	double	maxPos = 0.0;
	double	minNeg = 0.0;
	double	maxNeg = 0.0;
	double	posAvg = 0.0;
	double	negAvg = 0.0;
	double	testAcc = 0.0;
	double	trainAcc = 0.0;

	WMSet::PosIndices = sd.posIndices();
	WMSet::NegIndices = sd.negIndices();
	WMSet::NPositive = (int)sd.nPositive();
	WMSet::NNegative = (int)sd.nNegative();

	for(int i = 1; i <= maxSets; i++) {
		sd.addSubset(2000);
		if (i % dispInterval == 0) {
			testAcc = 0.0;
			trainAcc = sd.getAccuracy();
			if (nTest > 0) testAcc = sd.getAccuracy(nTest, testData, testClasses);

			sd.getDistValues(minPos, maxPos, posAvg, posSD, minNeg, maxNeg, negAvg, negSD);
			fprintf(stderr, "%d %.2f%% %.2f%% %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f\n",
					i, trainAcc*100.0, testAcc*100.0,
					minPos, posAvg, posSD, maxPos, minNeg, negAvg, negSD, maxNeg);
		}
	}
	if (dumpAccuracy) fprintf(stderr, "\n\n  --== Training Set Accuracy ==--\n");
	trainAcc = sd.getAccuracy(dumpAccuracy);

	if (dumpAccuracy && nTest) fprintf(stderr, "\n\n   --== Test Set Accuracy ==--\n");
	if (nTest) {
		testAcc = sd.getAccuracy(nTest, testData, testClasses, dumpAccuracy);
		fprintf(stderr, "\nTrainAcc = %.2f%% TestAcc = %.2f%%\n", trainAcc*100.0, testAcc*100.0);
	}
}

void
learnMultipleClassifier(SD **sdArray, int nClasses, int dispInterval,
						int nTrain, double **trainData, int *trainClasses, int **trainClassArray,
						int nTest, double **testData, int *testClasses, int **testClassArray, int dumpAccuracy)
{
	double	posSD = 0.0;
	double	negSD = 0.0;
	double	minPos = 0.0;
	double	maxPos = 0.0;
	double	minNeg = 0.0;
	double	maxNeg = 0.0;
	double	posAvg = 0.0;
	double	negAvg = 0.0;
	double	testAcc = 0.0;
	double	trainAcc = 0.0;
	int		maxSets = sdArray[0]->maxSets();

	for(int i = 1; i <= maxSets; i++) {
		multiAddSet(i, nClasses, sdArray);
		fprintf(stderr, ".");

		if ((i % dispInterval == 0) && i > 0) {
			fprintf(stderr, "\n");
			for(int j = 0; j < nClasses; j++) {
				SD		*sd = sdArray[j];

				testAcc = 0.0;
				trainAcc = sd->getAccuracy();

				if (nTest > 0) testAcc = sd->getAccuracy(nTest, testData, testClassArray[j]);
				sd->getDistValues(minPos, maxPos, posAvg, posSD, minNeg, maxNeg, negAvg, negSD);

				fprintf(stderr, "%5d %7.2f%% %7.2f%% %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f\n",
						j, trainAcc*100.0, testAcc*100.0, minPos, posAvg, posSD, maxPos, minNeg, negAvg, negSD, maxNeg);
			}
			if (dumpAccuracy && i == maxSets) fprintf(stderr, "\n\n  --== Training Set Accuracy ==--\n");
			trainAcc = getMultiAccuracy(sdArray, trainClasses, trainData, nClasses, nTrain, (dumpAccuracy && i == maxSets));

			if (nTest > 0) testAcc = getMultiAccuracy(sdArray, testClasses, testData, nClasses, nTest, (dumpAccuracy && i == maxSets));

			if (dumpAccuracy && nTest > 0 && i == maxSets) fprintf(stderr, "\n\n  --== Test Set Accuracy ==--\n");
			if (i == maxSets)
				fprintf(stderr, "\n\nTrainAcc = %.2f%%  TestAcc = %.2f%%\n", trainAcc*100.0, testAcc*100.0);
			else
				fprintf(stderr, "\n\n== %d %.2f%% %.2f%%\n", i, trainAcc*100.0, testAcc*100.0);
		}
	}
}

void
dumpBinaryClassifier(SD &sd, const char *dumpFileName)
{
	FILE	*dumpFile = fopen(dumpFileName, "w");

	sd.writeToFile(dumpFile);

	fclose(dumpFile);
}

void
dumpMultipleClassifiers(SD **sdArray, int nClasses, const char *dumpFileName)
{
	for(int i = 0; i < nClasses; i++) {
		char	fname[512];
		sprintf(fname, "%s.%d", dumpFileName, i);
		dumpBinaryClassifier(*sdArray[i], fname);
	}
}

void
loadData(const char *fname, int ndata, int ninputs, double **data, int *classes, DPair *bounds)
{
	int		i = 0;
	int		j = 0;
	FILE	*dataFile = fopen(fname, "r");

	if (dataFile == NULL) {
		perror(fname);
		exit(1);
	}

	for(i = 0; i < ndata; i++) {
		for(j = 0; j < ninputs; j++) {
			double	val = 0.0;

			fscanf(dataFile, "%lf", &val);
			data[i][j] = val;

			if (bounds != (DPair *)0) {
				DPair	&dp = bounds[j];
				dp.v0 = (dp.v0 < val) ? dp.v0 : val;
				dp.v1 = (dp.v1 > val) ? dp.v1 : val;
			}
		}
		fscanf(dataFile, "%d", &classes[i]);
	}

	fclose(dataFile);
}

double
getMultiAccuracy(SD **sdArray, int *classes, double **data, int nClasses, int nData, int doPrint)
{
	int		i = 0;
	int		j = 0;
	double	nDK = 0.0;
	double	nRight = 0.0;
	int		rights[20][20];

	if (doPrint) {
		for(i=0; i < 20; i++) {
			for(j=0; j < 20; j++)
				rights[i][j] = 0;
		}
		fprintf(stderr, "\n  index     ");
		for(j = 0; j < nClasses; j++)
			fprintf(stderr, "score%02d ", j);
		fprintf(stderr, "   target output\n");
	}

	for(i = 0; i < nData; i++) {
		int		classID = -1;
		double	maxAvg = -1.0e99;
		double	*point = data[i];

		if (doPrint) fprintf(stderr, "%010d ", i);
		for(j = 0; j < nClasses; j++) {
			double	avg = 0;
			SD		*sd = sdArray[j];

			sd->classifyPoint(point, avg);
			if (avg > maxAvg) {
				maxAvg = avg;
				classID = j;
			}
			if (doPrint) fprintf(stderr, "%7.2f ", avg);
		}
		if (classID == -1)	{
			classID = nClasses;
			nDK++;
		}
		rights[classes[i]][classID]++;
		if (classID == classes[i]) nRight++;
		if (doPrint) fprintf(stderr, "  %6d %6d\n", classes[i], classID);
	}
	if (doPrint) {
		fprintf(stderr, "\n          ");
		for(i = 0; i < nClasses; i++)
			fprintf(stderr, "     %d", i);
		fprintf(stderr, "  <- classified as\n");
		fprintf(stderr, "targets:");
		for(i = 0; i < nClasses; i++) {
			if (i > 0)
				fprintf(stderr, "        ");
			fprintf(stderr, "%2d ", i);
			for(j = 0; j < nClasses; j++)
				fprintf(stderr, "%5d ", rights[i][j]);
			fprintf(stderr, "\n");
		}
	}

	return nRight/nData;
}

void
multiAddSet(int nSets, int nClasses, SD **sdArray)
{
	int		i = 0;
	int		setsAdded = 0;
	int		setCounter = 0;
	int		nonUniform = 0;
	int		nonEnriched = 0;
	int		setsToAdd = nClasses;
	int		nDims = sdArray[0]->nDims();
	int		*classNonUniform = new int [nClasses];
	int		*classNonEnriched = new int [nClasses];
	double	kappa = sdArray[0]->kappa();

	for(i = 0; i < nClasses; i++) classNonEnriched[i] = classNonUniform[i] = 0;

	while(setsAdded < nClasses) {
		int		nModels = (int)(1.0 + kappa * drand48());

		for(i = 0; i < nClasses; i++) {
			int		cnu = classNonUniform[i];
			int		cni = classNonEnriched[i];
			SD		*sd = sdArray[i];
			WMSet	*newSet;

			if (sd->nSets() < nSets) {
				WMSet::PosIndices = sd->posIndices();
				WMSet::NegIndices = sd->negIndices();
				WMSet::NPositive = (int)sd->nPositive();
				WMSet::NNegative = (int)sd->nNegative();
				newSet = new WMSet(nDims, nModels);

				int			noEnrichment;
				int			noUniformity;
				double		fracPos;
				double		fracNeg;
				if (sd->addSubset(*newSet, noEnrichment, noUniformity, cni>4000)) {
					sd->takeNewSet(new WMSet(*newSet));
					setsAdded++;
					//fprintf(stderr, "class %d added set %d (%d ne/%d nu) ", i, nSets, cni, cnu);
					//newSet->print(stderr);
					setsToAdd--;
				}
				else
					delete newSet;

				classNonUniform[i] += noUniformity;
				classNonEnriched[i] += noEnrichment;
			}
		}
		setCounter++;

		if ((setCounter%1000) == 0) {
			fprintf(stderr, "%d sets to add after %d trial sets failed (at %d sets)\n",
					setsToAdd, setCounter, nSets);
			for(i = 0; i < nClasses; i++) {
				SD		*sd = sdArray[i];
				fprintf(stderr, "%c %d: %d %d\n",
						(sd->nSets()==nSets)?'*':' ',
						i, classNonEnriched[i], classNonUniform[i]);
			}
		}
	}
	delete [] classNonUniform;
	delete [] classNonEnriched;
}

#if 0
void
multiAddSet(int nSets, int nClasses, SD **sdArray)
{
	int		i = 0;
	int		setsAdded = 0;
	int		setCounter = 0;
	int		nonUniform = 0;
	int		nonEnriched = 0;
	int		setsToAdd = nClasses;
	int		nDims = sdArray[0]->nDims();
	int		*classNonUniform = new int [nClasses];
	int		*classNonEnriched = new int [nClasses];
	double	kappa = sdArray[0]->kappa();

	for(i = 0; i < nClasses; i++) classNonEnriched[i] = classNonUniform[i] = 0;

	while(setsAdded < nClasses) {
		int		nModels = (int)(1.0 + kappa * drand48());
		WMSet	*newSet = new WMSet(nDims, nModels);

		for(i = 0; i < nClasses; i++) {
			int		cnu = classNonUniform[i];
			int		cni = classNonEnriched[i];
			SD		*sd = sdArray[i];

			if (sd->nSets() < nSets) {
				int			noEnrichment;
				int			noUniformity;
				double		fracPos;
				double		fracNeg;
				if (sd->addSubset(*newSet, noEnrichment, noUniformity, cni>4000)) {
					sd->takeNewSet(new WMSet(*newSet));
					setsAdded++;
					fprintf(stderr, "class %d added set %d (%d ne/%d nu) ", i, nSets, cni, cnu);
					newSet->print(stderr);
					setsToAdd--;
				}
				classNonUniform[i] += noUniformity;
				classNonEnriched[i] += noEnrichment;
			}
		}
		delete newSet;
		setCounter++;

		if ((setCounter%1000) == 0) {
			fprintf(stderr, "%d sets to add after %d trial sets failed (at %d sets)\n",
					setsToAdd, setCounter, nSets);
			for(i = 0; i < nClasses; i++) {
				SD		*sd = sdArray[i];
				fprintf(stderr, "%c %d: %d %d\n",
						(sd->nSets()==nSets)?'*':' ',
						i, classNonEnriched[i], classNonUniform[i]);
			}
		}
	}
	delete [] classNonUniform;
	delete [] classNonEnriched;
}
#endif

