/*
	File: WMSet.C
	Component of the Stochastic Discrimination Utilities
	Copyright (C) 2002, David C. Lambert <dcl@panix.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation, Inc.,
	59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include <string.h>
#include "WMSet.H"
#include "RectWM.H"
#include "HammWM.H"
#include "PlaneWM.H"

typedef struct {
	WMSet::TWMCopier	copier;
	WMSet::TWMLoader	loader;
	WMSet::TWMFactory	factory;
	void				(*paramUpdate)();
} TWMFuncs;

static TWMFuncs WMFuncs[] = {
	{ RectWMCopier, RectWMLoader, RectWMFactory, RectWMParamUpdate },
	{ HammWMCopier, HammWMLoader, HammWMFactory, HammWMParamUpdate },
	{ PlaneWMCopier, PlaneWMLoader, PlaneWMFactory, PlaneWMParamUpdate },
	{ 0, 0, 0, 0 }
};

int					WMSet::NTrain;
int					WMSet::NPositive;
int					WMSet::NNegative;
int					*WMSet::TrainClasses;
const int			*WMSet::NegIndices;
const int			*WMSet::PosIndices;

double				WMSet::Tau;
double				WMSet::RhoMin;
double				WMSet::RhoMax;
double				WMSet::RhoDecay;
double				WMSet::SubsetThresh;
double				**WMSet::TrainData;

DPair				*WMSet::TrainBounds;

WMSet::TWMType		WMSet::WMType;
WMSet::TWMCopier	WMSet::WMCopier;
WMSet::TWMLoader	WMSet::WMLoader;
WMSet::TWMFactory	WMSet::WMFactory;
void				(*WMSet::WMParamUpdate)();

WeakModel::WeakModel(int nDims)
		: _nDims(nDims), _setDims(_nDims), _inSet(0)
{
	if (WMSet::SubsetThresh > 0.0)	{
		int		counter = 0;

		//_setDims = int(((drand48() * 0.5) + 0.5) * (1.0 - WMSet::SubsetThresh) * _nDims);
		_setDims = int((1.0 - WMSet::SubsetThresh) * _nDims);

		if (_setDims < 1)
			_setDims = 1;
		else if (_setDims > _nDims)
			_setDims = _nDims;
		

		_inSet = new int [_setDims];
		while(counter < _setDims)
			_inSet[counter++] = int(drand48() * _nDims);
	}
	//fprintf(stderr, "%d %d\n", _nDims, _setDims);
}

WeakModel::WeakModel(const WeakModel &rhs)
		: _nDims(rhs._nDims), _setDims(rhs._setDims), _inSet(0)
{
	if (rhs._inSet)	{
		_inSet = new int [_setDims];
		memcpy(_inSet, rhs._inSet, sizeof(int)*_setDims);
	}
}

WeakModel::WeakModel(int nDims, FILE *dumpFile)
		:_nDims(nDims), _setDims(0), _inSet(0)
{
	fscanf(dumpFile, "| %d ", &_setDims);

	if (_setDims < _nDims) {
		_inSet = new int [_setDims];

		for(int i = 0; i < _setDims; i++)
			fscanf(dumpFile, "%d ", &( _inSet[i]));
	}
	fscanf(dumpFile, " | ");
}

void
WeakModel::writeToFile(FILE *dumpFile)
{
	fprintf(dumpFile, "| %d ", _setDims);

	if (_setDims < _nDims)	{
		for(int i = 0; i < _setDims; i++)
			fprintf(dumpFile, "%d ", _inSet[i]);
	}
	fprintf(dumpFile, "|");
}

void
WeakModel::print(FILE *dumpFile)
{
}

double *
WMSet::RandomPositiveDataPoint()
{
	int		pointIndex = int(drand48() * WMSet::NPositive);

	return WMSet::TrainData[WMSet::PosIndices[pointIndex]];
}

double *
WMSet::RandomNegativeDataPoint()
{
	int		pointIndex = int(drand48() * WMSet::NNegative);

	return WMSet::TrainData[WMSet::NegIndices[pointIndex]];
}

double *
WMSet::RandomDataPoint(int *classID)
{
	int		pointIndex = int(drand48() * WMSet::NTrain);

	if (classID != (int *)0) *classID = WMSet::TrainClasses[pointIndex];

	return WMSet::TrainData[pointIndex];
}

int
WMSet::Initialize(int nTrain, double **trainData, int *trainClasses, DPair *bounds, TWMType wmType)
{
	TWMFuncs		&twmf = WMFuncs[wmType];

	if (twmf.copier == 0) return 0;

	WMSet::NTrain = nTrain;
	WMSet::TrainData = trainData;
	WMSet::TrainBounds = bounds;
	WMSet::TrainClasses = trainClasses;

	WMSet::WMType = wmType;
	WMSet::WMCopier = twmf.copier;
	WMSet::WMLoader = twmf.loader;
	WMSet::WMFactory = twmf.factory;
	WMSet::WMParamUpdate = twmf.paramUpdate;

	return 1;
}

WMSet::WMSet(int nDims, int nModels)
		: _nDims(nDims), _nModels(nModels), _fracPos(0.0), _fracNeg(0.0), _models(new WMPtr[nModels])
{
	for(int i=0; i < _nModels;i++)
		_models[i] = WMSet::WMFactory(*this, _nDims);
}

WMSet::WMSet(const WMSet &rhs)
		: _nDims(rhs._nDims), _nModels(rhs._nModels), _fracPos(rhs._fracPos), _fracNeg(rhs._fracNeg), _models(new WMPtr[rhs._nModels])
{
	for(int i=0; i < _nModels; i++)
		_models[i] = WMCopier(*(rhs._models[i]));
}

WMSet::WMSet(int nDims, FILE *dumpFile)
		: _nDims(nDims), _nModels(0),  _fracPos(0.0), _fracNeg(0.0), _models(0)
{
	int		mType;

	fscanf(dumpFile, "%d %d %lf %lf [ ", &_nModels, &mType, &_fracPos, &_fracNeg);
	_models = new WMPtr[_nModels];

	for(int i = 0; i < _nModels; i++)
		_models[i] = WMFuncs[mType].loader(_nDims, dumpFile);

	fscanf(dumpFile, "]");
}

int
WMSet::containsPoint(double *p)
{
	for(int i = 0; i < _nModels; i++) {
		WeakModel	*model = _models[i];
		if (!model->containsPoint(p)) return 0;
	}
	return 1;
}

void
WMSet::writeToFile(FILE *dumpFile)
{
	fprintf(dumpFile, "%d %d %.5e %.5e [ ", _nModels, (int)WMSet::WMType, _fracPos, _fracNeg);
	for(int i = 0; i < _nModels; i++) {
		WeakModel	*model = _models[i];
		model->writeToFile(dumpFile);
	}
	fprintf(dumpFile, "]\n");
}

void
WMSet::print(FILE *printFile)
{
	for(int i = 0; i < _nModels; i++)	{
		_models[i]->print(printFile);
	    if (_nModels > 1 && i+1 < _nModels) fprintf(printFile, ",");
	}
	fprintf(printFile, "\n");
}
