/*
    File: SDSet.H
    Component of the Stochastic Discrimination Utilities
    Copyright (C) 2002, David C. Lambert <dcl@panix.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#ifndef _SDSET_H
#define _SDSET_H

#include <stdio.h>

class WMSet;
class WeakModel;

struct DPair {
	DPair() : v0(0.0), v1(0.0) {}
	DPair(double vv0, double vv1) : v0(vv0), v1(vv1) {}
	DPair(const DPair &rhs) : v0(rhs.v0), v1(rhs.v1) {}
	DPair	&operator=(const DPair &rhs) { if (&rhs == this) return *this; v0 = rhs.v0; v1 = rhs.v1; } 
	double	v0;
	double	v1;
};

class WeakModel {
  public:
	WeakModel(int nDims);
	WeakModel(int nDims, FILE *dumpFile);
	WeakModel(const WeakModel &rhs);

	virtual ~WeakModel() { delete [] _inSet; }

	int				nDims() { return _nDims; }
	int				setDims() { return _setDims; }
	virtual int		containsPoint(double *p) { return 0; }
	virtual void	writeToFile(FILE *dumpFile);
	virtual void	print(FILE *printFile);

  protected:
	int		_nDims;
	int		_setDims;
	int		*_inSet;
};

class WMSet {
  public:
	typedef enum { INVALID_WMTYPE=-1, HRECT, HAMM, HPLANE}	TWMType;

	typedef WeakModel	*WMPtr;
	typedef WMPtr		(*TWMLoader)(int, FILE *);
	typedef WMPtr		(*TWMCopier)(const WeakModel &);
	typedef WMPtr		(*TWMFactory)(WMSet &, int);


	static int			NTrain;
	static int			NPositive;
	static int			NNegative;
	static const int	*PosIndices;
	static const int	*NegIndices;

	static double		Tau;
	static double		RhoMin;
	static double		RhoMax;
	static double		RhoDecay;
	static double		SubsetThresh;

	static void			UpdateWMParams();
	static void			(*WMParamUpdate)();
	static int			Initialize(int, double **, int *, DPair *, TWMType);

	static DPair		&DataBounds(int i) { return WMSet::TrainBounds[i]; }

	static double		*RandomDataPoint(int *cid=(int*)0);
	static double		*RandomPositiveDataPoint();
	static double		*RandomNegativeDataPoint();

	WMSet(const WMSet &rhs);
	WMSet(int nDims, int nModels);
	WMSet(int nDims, FILE *dumpFile);

	virtual ~WMSet()	{ for(int i=0; i<_nModels;i++) delete _models[i]; delete [] _models ; }

	int			nDims() { return _nDims; }
	int			nModels() { return _nModels; }

	double		fracPos() { return _fracPos; }
	void		fracPos(double v) { _fracPos = v; }

	double		fracNeg() { return _fracNeg; }
	void		fracNeg(double v) { _fracNeg = v; }

	WeakModel	*model(int i)  { return _models[i]; }
	WeakModel	*setModel(int i, WeakModel *newModel) { WeakModel *oldModel = _models[i]; _models[i] = newModel; return oldModel; }

	int			containsPoint(double *p);
	void		writeToFile(FILE *dumpFile);
	void		print(FILE *printFile);

  protected:
	static int			*TrainClasses;
	static double		**TrainData;
	static DPair		*TrainBounds;
	static TWMType		WMType;
	static TWMCopier	WMCopier;
	static TWMLoader	WMLoader;
	static TWMFactory	WMFactory;
	static TWMType	   	ModelType;

	int					_nDims;
	int					_nModels;
	double				_fracPos;
	double				_fracNeg;
	WeakModel			**_models;
};


#endif
